/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
#include "jsm.h"

#include "radiusclient.h"

#define AUTH_PASS 0
#define AUTH_REJECT 1
#define AUTH_FAILED 2

int mod_auth_radius_cpw(char *uname, char *pw)
{
  char		username[128];
  char		passwd[AUTH_PASS_LEN + 1];
  VALUE_PAIR	*send, *received;
  UINT4		service;
  char		msg[4096];
  int		result;

  strncpy(username, uname, sizeof(username));
  strncpy(passwd, pw, sizeof(passwd));

  send = NULL;

  if (rc_avpair_add(&send, PW_USER_NAME, username, 0) == NULL)
  {
    log_error("mod_auth_radius","Adding username %s", username);
    log_error("mod_auth_radius","Failed adding username failed.");
    return AUTH_FAILED;
  }

  if (rc_avpair_add(&send, PW_USER_PASSWORD, passwd, 0) == NULL)
  {
    rc_avpair_free(send);
    log_error("mod_auth_radius","Failed adding password failed.");
    return AUTH_FAILED;
  }

  service = PW_AUTHENTICATE_ONLY;
  if (rc_avpair_add(&send, PW_SERVICE_TYPE, &service, 0) == NULL)
  {
    rc_avpair_free(send);
    log_error("mod_auth_radius","Failed adding service type.");
    return AUTH_FAILED;
  }

  log_debug("mod_auth_radius","sending request to RADIUS");

  result = rc_auth(0, send, &received, msg);
  rc_avpair_free(send);

  switch (result)
  {
  case OK_RC:
    return AUTH_PASS;
    break;

  case BADRESP_RC:
    return AUTH_REJECT;
    break;

  default:
    return AUTH_FAILED;
    break;
  }

    
}

mreturn mod_auth_radius_go(mapi m, void *arg)
{
  char	*pass;

  log_debug("mod_auth_radius","checking");

  if(jpacket_subtype(m->packet) == JPACKET__GET)
  { /* type=get means we flag that the server can do plain-text auth */
    xmlnode_insert_tag(m->packet->iq,"password");
    return M_PASS;
  }

  if((pass = xmlnode_get_tag_data(m->packet->iq, "password")) == NULL)
    return M_PASS;

  switch(mod_auth_radius_cpw(m->user->user, pass))
  {
  case AUTH_PASS:
    jutil_iqresult(m->packet->x);
    break;

  case AUTH_REJECT:
    jutil_error(m->packet->x, TERROR_AUTH);
    break;

  default:
    return AUTH_FAILED;
  }
  return M_HANDLED;
}

/* handle registration of a new password */
mreturn mod_auth_radius_reg(mapi m, void *arg)
{
  char *pw;
  xmlnode q, reg;

  /* If config has no register section, we don't register anyone */
  if ((reg = js_config(m->si, "register")) == NULL) return M_PASS;

  log_debug("mod_auth_radius","registering");

  switch(jpacket_subtype(m->packet))
  {
  case JPACKET__GET:
    /* we are being asked for the required registration info */
    jutil_iqresult(m->packet->x);

    /* create a new query? */
    q = xmlnode_insert_tag(m->packet->x, "query");
    xmlnode_put_attrib(q,"xmlns",NS_REGISTER);

    /* copy the registration fields from the config file */
    xmlnode_insert_node(q,xmlnode_get_firstchild(reg));

    break;

  case JPACKET__SET:
    log_debug("mod_auth_radius","processing valid registration for %s",jid_full(m->packet->to));

    /* If there is no password in the request, barf */
    if ((pw = xmlnode_get_tag_data(m->packet->iq, "password")) == NULL)
    {
      jutil_error(m->packet->x,(terror){500,"No Password in Request"});
      break;
    }

    /* If the password is not correct according to RADIUS, barf */
    if (mod_auth_radius_cpw(xmlnode_get_tag_data(m->packet->iq,"username"), pw) != AUTH_PASS)
    {
      log_debug("mod_auth_radius","check PW for register failed");
      jutil_error(m->packet->x,TERROR_FORBIDDEN);
      break;
    }

    /* Lookin' good so far */

    /* Why should we have to do this? */
    /* users.c requires there be some kind of auth data */
    q = xmlnode_new_tag("password");
    xmlnode_insert_cdata(q,"**RADIUS**",-1);
    xdb_set(m->si->xc,jid_user(m->packet->to),NS_AUTH,q);

    /* We don't want username/password info in reg db */
    xmlnode_hide(xmlnode_get_tag(m->packet->iq,"password"));
    xmlnode_hide(xmlnode_get_tag(m->packet->iq,"username"));

    /* ??? */
    jutil_delay(m->packet->iq,"registered");
    
    xdb_set(m->si->xc,jid_user(m->packet->to),NS_REGISTER,m->packet->iq);

    /* send admins notice */
    if(xmlnode_get_attrib(reg,"notify") != NULL)
    {
      q = jutil_msgnew(NULL,m->packet->to->server,"Registration Notice",
		       spools(m->packet->p,"The user ",jid_full(m->packet->to),
			      " was just created with the following registration data: ",xmlnode2str(m->packet->iq),m->packet->p));
      xmlnode_put_attrib(q,"from",m->packet->to->server);
      js_deliver(m->si,jpacket_new(q));
    }

    /* send the user a welcome message */
    if((reg = js_config(m->si,"welcome")) != NULL)
    {
      q = xmlnode_new_tag("message");
      xmlnode_put_attrib(q, "from", m->packet->to->server);
      xmlnode_put_attrib(q, "to", jid_full(m->packet->to));
      xmlnode_insert_node(q,xmlnode_get_firstchild(reg));
      js_deliver(m->si,jpacket_new(q));
    }

    /* respond */
    jutil_iqresult(m->packet->x);
    break;

  default:
    return M_PASS;
  }

  return M_HANDLED;

}

/* handle password change requests from a session */
mreturn mod_auth_radius_server(mapi m, void *arg)
{
  xmlnode reg, q, cur, check, pass;

  /* pre-requisites */
  if(m->packet->type != JPACKET_IQ) return M_IGNORE;
  if(!NSCHECK(m->packet->iq,NS_REGISTER)) return M_PASS;
  if(m->user == NULL) return M_PASS;
  if(js_config(m->si,"register") == NULL) return M_PASS;

  log_debug("mod_auth_radius","updating server: %s, user %s",m->user->id->server,jid_full(m->user->id));

  /* check for their registration */
  reg = xdb_get(m->si->xc,m->user->id,NS_REGISTER);

  switch(jpacket_subtype(m->packet))
  {
  case JPACKET__GET:
    /* create reply to the get */
    jutil_iqresult(m->packet->x);
    
    /* create a new query */
    q = xmlnode_insert_tag(m->packet->x,"query");

    xmlnode_put_attrib(q,"xmlns",NS_REGISTER);
    /* Do we really want to insert this? We don't allow password changes */
    xmlnode_insert_tag(q,"password");

    /* copy the registration fields from the config file */
    xmlnode_insert_node(q,xmlnode_get_firstchild(js_config(m->si,"register")));
    
    /* insert the key, we don't need to check it, but we'll send it */
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,"foobar"),-1);

    /* replace fields with already-registered ones */
    for(cur = xmlnode_get_firstchild(q);cur != NULL;cur = xmlnode_get_nextsibling(cur))
    {
      if(xmlnode_get_type(cur) != NTYPE_TAG) continue;

      check = xmlnode_get_tag(reg,xmlnode_get_name(cur));
      if(check == NULL) continue;

      xmlnode_insert_node(cur,xmlnode_get_firstchild(check));
    }

    /* add the registered flag and hide the username flag */
    xmlnode_insert_tag(q,"registered");
    
    break;

  case JPACKET__SET:
    if (xmlnode_get_tag(m->packet->iq,"remove") != NULL)
    {
      log_notice(m->user->id->server,"User Unregistered: %s",m->user->user);

      /* XXX BRUTE FORCE: remove the registration and auth and any misc data */
      xdb_set(m->si->xc,m->user->id,NS_REGISTER,NULL);
      xdb_set(m->si->xc,m->user->id,NS_AUTH,NULL);
      xdb_set(m->si->xc,m->user->id,NS_PRIVATE,NULL);
      xdb_set(m->si->xc,m->user->id,NS_ROSTER,NULL);
      xdb_set(m->si->xc,m->user->id,NS_VCARD,NULL);
      xdb_set(m->si->xc,m->user->id,NS_OFFLINE,NULL);
      xdb_set(m->si->xc,m->user->id,NS_FILTER,NULL);
    } else {
      log_debug("mod_auth_radius","updating registration for %s",jid_full(m->user->id));

      /* notify if password change is attempted */
      if((pass = xmlnode_get_tag(m->packet->iq,"password")) != NULL)
      {
	q = jutil_msgnew(NULL,jid_full(m->packet->to),"Password Change",
			 spools(m->packet->p,"The password for ",jid_full(m->packet->to)," can not be changed, because this system does not supporty password changes."));
	xmlnode_put_attrib(q,"from",m->packet->to->server);
	js_deliver(m->si,jpacket_new(q));
      }

      /* update the registration data */
      xmlnode_hide(xmlnode_get_tag(m->packet->iq,"username"));
      xmlnode_hide(xmlnode_get_tag(m->packet->iq,"password"));
      jutil_delay(m->packet->iq,"updated");
      xdb_set(m->si->xc,m->user->id,NS_REGISTER,m->packet->iq);
    }

    /* clean up and respond */
    jutil_iqresult(m->packet->x);
    break;

  default:
    xmlnode_free(reg);
    return M_PASS;

  }

  xmlnode_free(reg);
  js_deliver(m->si, jpacket_reset(m->packet));
  return M_HANDLED;

}

void mod_auth_radius(jsmi si)
{
  xmlnode q;

  log_debug("mod_auth_radius","init");

  if ((q = js_config(si, "radiusclient")) == NULL)
  {
    log_error("mod_auth_radius","radiusclient config not set");
    return;
  }

  rc_openlog("radiuslog");

  log_debug("mod_auth_radius", "radius config : %s",xmlnode_get_tag_data(q, "config"));

  if (rc_read_config(xmlnode_get_tag_data(q, "config")) != 0)
  {
    log_error("mod_auth_radius", "unable to load radius config");
    return;
  }

  if (rc_read_dictionary(rc_conf_str("dictionary")) != 0)
  {
    log_error("mod_auth_radius", "unable to load radius dictionary");
    return;
  }

  js_mapi_register(si, e_AUTH, mod_auth_radius_go, NULL);
  js_mapi_register(si, e_SERVER, mod_auth_radius_server, NULL);
  js_mapi_register(si, e_REGISTER, mod_auth_radius_reg, NULL);
}
